/*
 *    WindowClassificationPerformanceEvaluator.java
 *    Copyright (C) 2009 University of Waikato, Hamilton, New Zealand
 *    @author Albert Bifet (abifet@cs.waikato.ac.nz)
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program. If not, see <http://www.gnu.org/licenses/>.
 *    
 */
package moa.evaluation;

import java.util.LinkedList;
import java.util.TreeSet;

import moa.core.Measurement;
import moa.core.ObjectRepository;
import moa.options.AbstractOptionHandler;
import moa.options.IntOption;
import moa.tasks.TaskMonitor;
import weka.core.Utils;
import weka.core.Instance;

/**
 * Classification evaluator that updates diversity evaluation using a sliding
 * window. Only to be used for binary classification problems with unweighted instances.
 * 
 * @author Dariusz Brzezinski (dariusz.brzezinski at cs.put.poznan.pl)
 */
public class DiversityPerformanceEvaluator extends
		AbstractOptionHandler implements ClassificationPerformanceEvaluator {

	private static final long serialVersionUID = 1L;

	private int widthOption;

	protected double totalObservedInstances = 0;

	private Estimator diversityEstimator;

	protected int numClasses;

	public class Estimator {

		public class Score implements Comparable<Score> {
			/**
			 * Predicted score of the example
			 */
			protected double value;

			/**
			 * Age of example - position in the window where the example was
			 * added
			 */
			protected int posWindow;

			/**
			 * True if example's true label is positive
			 */
			protected boolean isPositive;

			/**
			 * Constructor.
			 * 
			 * @param value
			 *            score value
			 * @param position
			 *            score position in window (defines its age)
			 * @param isPositive
			 *            true if the example's true label is positive
			 */
			public Score(double value, int position, boolean isPositive) {
				this.value = value;
				this.posWindow = position;
				this.isPositive = isPositive;
			}

			/**
			 * Sort descending based on score value.
			 */
			@Override
			public int compareTo(Score o) {
				if (o.value < this.value) {
					return -1;
				} else if (o.value > this.value){
					return 1;
				} else {
					if (o.posWindow > this.posWindow) {
						return -1;
					} else if (o.posWindow < this.posWindow){
						return 1;
					} else {
						return 0;
					}
				}
			}
			
			@Override
			public boolean equals(Object o) {
				return (o instanceof Score) && ((Score)o).posWindow == this.posWindow;
			}
		}

		protected TreeSet<Score> sortedScores;	
		protected TreeSet<Score> holdoutSortedScores;
		protected Score[] scoreWindow;
		protected int numClasses;
		protected int ensembleSize;

		protected double[] ensemblePredictions;
		protected double[][] componentPredictions;
		protected int posWindow;
		protected int size;
		
		protected double numPos;
		protected double numNeg;
		protected double holdoutNumPos;
		protected double holdoutNumNeg;
		
		protected double correctPredictions;
		
	    protected double[] columnKappa;
	    protected double[] rowKappa;
	    protected int[] predictedClasses;
	    protected int[] trueClasses;
	    

		public Estimator(int sizeWindow, int numClasses) {
			this.numClasses = numClasses;
			this.sortedScores = new TreeSet<Score>();
			this.holdoutSortedScores = new TreeSet<Score>();
			this.size = sizeWindow;
			this.scoreWindow = new Score[sizeWindow];
			this.ensemblePredictions = new double[sizeWindow];
			this.componentPredictions = new double[sizeWindow][];
			this.trueClasses = new int[sizeWindow];
			this.predictedClasses = new int[sizeWindow];
			
	        this.rowKappa = new double[numClasses];
	        this.columnKappa = new double[numClasses];
	        for (int i = 0; i < numClasses; i++) {
	            this.rowKappa[i] = 0.0;
	            this.columnKappa[i] = 0.0;
	        }
			
			this.posWindow = 0;
			this.numPos = 0;
			this.numNeg = 0;
			this.holdoutNumPos = 0;
			this.holdoutNumNeg = 0;
			this.correctPredictions = 0;
		}

		protected void addBinaryStats(int trueClass, double[] classVotes){
			Double normalizedVote = 1.0;
			
			if(classVotes.length == 2) {
				normalizedVote = classVotes[0]/(classVotes[0] + classVotes[1]);
			}
			
			if(normalizedVote.isNaN()){
				normalizedVote = 0.0;
			}
			
			double score = normalizedVote;
			boolean isPositive = trueClass == 0; 
			
            // // periodically update holdout evaluation
			if (size > 0 && posWindow % this.size == 0) {
				this.holdoutSortedScores = new TreeSet<Score>();
				
				for (Score s : this.sortedScores) {
					this.holdoutSortedScores.add(s);
				}
				
				this.holdoutNumPos = this.numPos;
				this.holdoutNumNeg = this.numNeg;
			}
			
			if (size > 0 && posWindow >= this.size) {
				// // remove the oldest example
				sortedScores.remove(scoreWindow[posWindow % size]);
				
				if (scoreWindow[posWindow % size].isPositive) {
					numPos--;
				} else {
					numNeg--;
				}
			}
			
			Score newScore = new Score(score, posWindow, isPositive);
			sortedScores.add(newScore);
			if (newScore.isPositive) {
				numPos++;
			} else {
				numNeg++;
			}
			
			if (size > 0) {
				scoreWindow[posWindow % size] = newScore;
			}
		}
		
		public void add(int trueClass, double[] overallVotes, double[][] classVotes) {
			int predictedClass = Utils.maxIndex(overallVotes);
			boolean correctPrediction = predictedClass == trueClass;
			this.ensembleSize = classVotes.length;
			
			if(this.numClasses == 2) {
				addBinaryStats(trueClass, overallVotes);
			}
					
			if (size > 0 && posWindow >= this.size) {
				correctPredictions -= ensemblePredictions[posWindow % size];
				
				int oldestExampleTrueClass = trueClasses[posWindow % size];
	            int oldestExamplePredictedClass = predictedClasses[posWindow % size];  
	            
				this.rowKappa[oldestExamplePredictedClass] -= 1;
	            this.columnKappa[oldestExampleTrueClass] -= 1;
			}
			
			// // add new example
			correctPredictions += correctPrediction ? 1 : 0;
			
            this.rowKappa[predictedClass] += 1;
            this.columnKappa[trueClass] += 1;

			if (size > 0) {
				ensemblePredictions[posWindow % size] = correctPrediction ? 1 : 0;
				componentPredictions[posWindow % size] = getComponentPredictions(trueClass, classVotes);
				trueClasses[posWindow % size] = trueClass;
				predictedClasses[posWindow % size] = predictedClass;
			}
			
			//// posWindow needs to be always incremented to differentiate between examples in the red-black tree
			posWindow++;
		}

		private double[] getComponentPredictions(int trueClass, double[][] classVotes){
			double[] instancePredictions = new double[classVotes.length];
			
			for (int i=0; i<classVotes.length; i++) {			
				instancePredictions[i] = (Utils.maxIndex(classVotes[i]) == trueClass) ? 1 : 0;
			}
			
			return instancePredictions;
		}
		
		public double getAUC() {
			double AUC = 0;
			double c = 0;
			
			if (numPos == 0 || numNeg == 0) {
				return 1;
			}

			for (Score s : sortedScores){
				if(s.isPositive) {
					c += 1;
				} else {
					AUC += c;
				}
			}
			
			return AUC / (numPos * numNeg);
		}
		
		public double getHoldoutAUC() {
			double AUC = 0;
			double c = 0;

			if (holdoutSortedScores.isEmpty()) {
				return 0;
			}
			
			if (holdoutNumPos == 0 || holdoutNumNeg == 0) {
				return 1;
			}

			for (Score s : holdoutSortedScores){
				if(s.isPositive) {
					c += 1;
				} else {
					AUC += c;
				}
			}
			
			return AUC / (holdoutNumPos * holdoutNumNeg);
		}

		public double getScoredAUC() {
			double AOC = 0;
			double AUC = 0;
			double r = 0;
			double c = 0;
			double R_plus, R_minus;
			
			if (numPos == 0 || numNeg == 0) {
				return 1;
			}
			
			for (Score s : sortedScores){
				if(s.isPositive) {
					c += s.value;
					AOC += r;
				} else {
					r += s.value;
					AUC += c;
				}
			}
			
			R_minus = (numPos*r - AOC)/(numPos * numNeg);
			R_plus = (AUC)/(numPos * numNeg);		
			return R_plus - R_minus;
		}
		
		public double getRatio() {
			if(numNeg == 0) {
				return Double.MAX_VALUE;
			} else {
				return numPos/numNeg;
			}
		}
		
		public double getAccuracy() {
			if (size > 0) {
				return totalObservedInstances > 0.0 ? correctPredictions / Math.min(size, totalObservedInstances) : 0.0;
			} else {
				return totalObservedInstances > 0.0 ? correctPredictions / totalObservedInstances : 0.0;
			}
		}
		
		public double getKappa() {
            double p0 = getAccuracy();
            double pc = 0.0;
            
            if (size > 0) {
	            for (int i = 0; i < numClasses; i++) {
	                pc += (this.rowKappa[i]/Math.min(size, totalObservedInstances)) * (this.columnKappa[i]/Math.min(size, totalObservedInstances));
	            }
            } else {
            	for (int i = 0; i < numClasses; i++) {
	                pc += (this.rowKappa[i]/totalObservedInstances) * (this.columnKappa[i]/totalObservedInstances);
	            }
            }
            return (p0 - pc) / (1.0 - pc);
	    }
		
		/////////////////////////////////////
		/// Diversity measures
		/////////////////////////////////////
		
		private double diversityContingency(int i, int j, double expectedi, double expectedj){
			double value = 0.0;
			
			for (int pos = 0; pos < componentPredictions.length; pos++){
				if (componentPredictions[pos] != null && i < componentPredictions[pos].length && j < componentPredictions[pos].length ){
					value += (componentPredictions[pos][i] == expectedi && componentPredictions[pos][j] == expectedj) ? 1 : 0;
				}
			}
			
			if (size > 0) {
				return totalObservedInstances > 0.0 ? value / Math.min(size, totalObservedInstances) : 0.0;
			} else {
				return totalObservedInstances > 0.0 ? value / totalObservedInstances : 0.0;
			}
		}
		
		private double a(int i, int j) {
			return this.diversityContingency(i, j, 1, 1);
		}
		
		private double b(int i, int j) {
			return this.diversityContingency(i, j, 1, 0);
		}
		
		private double c(int i, int j) {
			return this.diversityContingency(i, j, 0, 1);
		}
		
		private double d(int i, int j) {
			return this.diversityContingency(i, j, 0, 0);
		}
		
		private double p(int i) {
			double pSum = 0.0;
			
			for (int pos = 0; pos < componentPredictions.length; pos++){
				int failCount = 0; 
				
				for (int c = 0; componentPredictions[pos] != null && c < componentPredictions[pos].length; c++){
					if (componentPredictions[pos][c] == 0){
						failCount++;
					}
				}
				
				if (failCount == i){
					pSum += 1.0;
				}
			}
			
			if (size > 0) {
				return totalObservedInstances > 0.0 ? pSum / Math.min(size, totalObservedInstances) : 0.0;
			} else {
				return totalObservedInstances > 0.0 ? pSum / totalObservedInstances : 0.0;
			}
		}
		
		private double getComponentAccuracy(int i) {
			double correctComponentPredictions = 0.0;
			
			for (int pos = 0; pos < componentPredictions.length; pos++){
				if (i < componentPredictions[pos].length){
					correctComponentPredictions += componentPredictions[pos][i];
				}
			}
			
			if (size > 0) {
				return totalObservedInstances > 0.0 ? correctComponentPredictions / Math.min(size, totalObservedInstances) : 0.0;
			} else {
				return totalObservedInstances > 0.0 ? correctComponentPredictions / totalObservedInstances : 0.0;
			}
		}
		
		public double getQStatistic() {
			double qCount = 0;
			double qSum = 0;
			
			for(int i = 0; i <this.ensembleSize-1; i++){
				for (int j = i+1; j <this.ensembleSize; j++){
					qSum += (a(i, j)*d(i, j) - b(i, j)*c(i, j))/
							(a(i, j)*d(i, j) + b(i, j)*c(i, j));
					qCount += 1;
				}
			}
			
			return qSum/qCount;
		}
		
		public double getDisagreement() {
			double dCount = 0;
			double dSum = 0;
			
			for(int i = 0; i <this.ensembleSize-1; i++){
				for (int j = i+1; j <this.ensembleSize; j++){
					dSum += b(i, j) + c(i, j);
					dCount += 1;
				}
			}
			
			return dSum/dCount;
		}
		
		public double getDoubleFault() {
			double dCount = 0;
			double dSum = 0;
			
			for(int i = 0; i <this.ensembleSize-1; i++){
				for (int j = i+1; j <this.ensembleSize; j++){
					dSum += d(i, j);
					dCount += 1;
				}
			}
			
			return dSum/dCount;
		}
		
		public double getKohaviWolpert() {
			return ((double)this.ensembleSize - 1.0)/(2.0*(double)this.ensembleSize) * this.getDisagreement();
		}
		
		public double getCFD() {
			double p_0 = p(0);
			
			if (p_0 == 1.0 || this.ensembleSize < 2){
				return 0;
			} else {
				double pSum = 0;
				
				for(int i = 1; i <= this.ensembleSize; i++){
					pSum += ((this.ensembleSize-i)/(this.ensembleSize-1))*p(i);
				}
				
				return 1/(1 - p_0)*pSum;
			}
		}
		
		public double getAgreementKappa() {
			double p = 0;
			
			for(int i = 0; i < this.ensembleSize; i++){
				p += this.getComponentAccuracy(i);
			}
			
			p = p / (double)this.ensembleSize;
			
			return 1 - 1/(2*p*(1-p)) * this.getDisagreement();
		}
	}

	public DiversityPerformanceEvaluator(int widthOption){
		this.widthOption = widthOption;
	}
	
	@Override
	public void reset() {
		reset(this.numClasses);
	}

	public void reset(int numClasses) {
		this.numClasses = numClasses;
		
		this.diversityEstimator = new Estimator(this.widthOption, this.numClasses);
		this.totalObservedInstances = 0;
	}

	@Override
	public void addResult(Instance inst, double[] overallVotes) {
		double weight = inst.weight();

		if (weight > 0.0) {
			// // initialize evaluator
			if (totalObservedInstances == 0) {
				reset(inst.dataset().numClasses());
			}
			this.totalObservedInstances += 1;
			
			this.diversityEstimator.add((int) inst.classValue(), overallVotes, null);
		}
	}
	
	public void addResult(Instance inst, double[] overallVotes, double[][] classVotes) {
		double weight = inst.weight();

		if (weight > 0.0) {
			// // initialize evaluator
			if (totalObservedInstances == 0) {
				reset(inst.dataset().numClasses());
			}
			this.totalObservedInstances += 1;
			
			this.diversityEstimator.add((int) inst.classValue(), overallVotes, classVotes);
		}
	}

	@Override
	public Measurement[] getPerformanceMeasurements() {
		return new Measurement[] {
				new Measurement("classified instances", this.totalObservedInstances),			
				new Measurement("AUC", this.diversityEstimator.getAUC()),
				new Measurement("sAUC", this.diversityEstimator.getScoredAUC()),
				new Measurement("Accuracy", this.diversityEstimator.getAccuracy()),
				new Measurement("Kappa", this.diversityEstimator.getKappa()),
				new Measurement("Periodical holdout AUC", this.diversityEstimator.getHoldoutAUC()),
				new Measurement("Pos/Neg ratio", this.diversityEstimator.getRatio()),
				// diversity measures
				new Measurement("Q-statistic", this.diversityEstimator.getQStatistic()),
				new Measurement("Disagreement", this.diversityEstimator.getDisagreement()),
				new Measurement("Double fault", this.diversityEstimator.getDoubleFault()),
				new Measurement("Kohavi-Wolpert variance", this.diversityEstimator.getKohaviWolpert()),
				new Measurement("Coincident Failure Diversity", this.diversityEstimator.getCFD()),
				new Measurement("Interrater Agreement Kappa", this.diversityEstimator.getAgreementKappa())
		};

	}

	@Override
	public void getDescription(StringBuilder sb, int indent) {
		Measurement.getMeasurementsDescription(getPerformanceMeasurements(),
				sb, indent);
	}

	@Override
	public void prepareForUseImpl(TaskMonitor monitor,
			ObjectRepository repository) {
	}

	public Estimator getEstimator() {
		return diversityEstimator;
	}
}
